#ifndef __CATALYZER_HPP
#define __CATALYZER_HPP 1

#include <mpi.h>
#include <iostream>
#include <fstream>
#include <cstring>

#include <config.h>

#include <vtkSmartPointer.h>
#include <vtkDataSet.h>
#include <vtkInformation.h>
#include <vtkMultiBlockDataSet.h>
#include <vtkXMLMultiBlockDataWriter.h>

/* Only use MPI when we have vtk 5.8+ */
/* features initializing MPI using an external context a missing in 5.8- */
/* but lets aim for the latest major version 6 to reduce the complexity */
#if VTK_MAJOR_VERSION >= 6 && defined(VTK_HAS_PARALLEL)
#include <vtkMPI.h>
#include <vtkMPIController.h>
#include <vtkMPICommunicator.h>
#include <vtkXMLPMultiBlockDataWriter.h>

#endif // VTK_MAJOR_VERSION >= 6 && defined(VTK_HAS_PARALLEL)

#include <vtkCPProcessor.h>
#include <vtkCPPipeline.h>
#include <vtkCPPythonScriptPipeline.h>

#include <vtkBaseInsituPipeline.h>

/**
 * \class Catalyzer
 * \brief Export to VTK format
 *
 * @author Alexandre Ancel
 */
class Catalyzer
{
public:

    /** @name Constructors, destructor
     */
    Catalyzer();
    Catalyzer( MPI_Comm comm );

    ~Catalyzer();

    /**
       Initialize Catalyst and VTK classes
     */
    void Initialize();

    /**
       save the timeset
     */
    void Save(int index, double time, vtkSmartPointer<vtkDataSet> out);

    /**
     * Actual write of the dataset into a file 
     */
    void Write( int stepIndex, std::string filename, vtkSmartPointer<vtkMultiBlockDataSet> out);

    void EnableSteering() { M_steeringEnabled = true; }
    void UsePythonPipeline( char * pythonScript ) { M_pythonPipeline = std::string(pythonScript); }
    void SetInsituHostname( std::string hostname ) { M_insituHostname = hostname; }
    void SetInsituPort( int port ) { M_insituPort = port; }

    /**
     * Attempt to get the steering value from the simulation
     * If no steering is enabled or the value cannot be accessed,
     * valid is set to false
     */
    void GetSteeringValue( double * v, int * invalid );

private:

    /**
     * Build a multi block structure based on the data gathered 
     * on the different processes.
     */
    vtkSmartPointer<vtkMultiBlockDataSet> BuildMultiBlockDataSet( double time, vtkSmartPointer<vtkDataSet> out );

    /* class members for in-situ visualization */
    MPI_Comm M_mpiComm;

#if VTK_MAJOR_VERSION >= 6 && defined(VTK_HAS_PARALLEL)
    vtkMPICommunicatorOpaqueComm * M_opaqueComm;
#endif
    vtkSmartPointer<vtkCPProcessor> M_inSituProcessor;

    vtkSMProxyManager* M_pxm;

    std::string M_pythonPipeline;

    std::string M_insituHostname;
    int M_insituPort;

    bool M_steeringEnabled;
    bool M_initialized;
};

#endif /* __Catalyzer_H */
