/**
   \file catalyzer.cpp
   \author Alexandre Ancel <alexandre.ancel@cemosis.fr>
   \date 2014-11-13
*/
#ifndef __CATALYZER_CPP
#define __CATALYZER_CPP 1

#include "catalyzer.h"

Catalyzer::Catalyzer()
:
M_inSituProcessor(NULL),
M_pxm(NULL),
#if VTK_MAJOR_VERSION >= 6 && defined(VTK_HAS_PARALLEL)
M_mpiComm(NULL),
M_opaqueComm(NULL),
#endif
M_initialized(false),
M_insituHostname("localhost"),
M_insituPort(22222),
M_steeringEnabled(false)
{ }

Catalyzer::Catalyzer( MPI_Comm comm )
:
M_inSituProcessor(NULL),
M_pxm(NULL),
#if VTK_MAJOR_VERSION >= 6 && defined(VTK_HAS_PARALLEL)
M_mpiComm(comm),
M_opaqueComm(NULL),
#endif
M_initialized(false),
M_insituHostname("localhost"),
M_insituPort(22222),
M_steeringEnabled(false)
{ }

Catalyzer::~Catalyzer()
{
#if VTK_MAJOR_VERSION >= 6 && defined(VTK_HAS_PARALLEL)
    /* end up in situ simulation */
    M_inSituProcessor->Finalize();

    if(M_opaqueComm)
    {
        /* clean memory */
        delete M_opaqueComm;
    }
#endif
}

void
Catalyzer::Initialize()
{
    std::cout << "Initializing Catalyst ..." << std::endl;

#if VTK_MAJOR_VERSION >= 6 && defined(VTK_HAS_PARALLEL)
    /* before version 5.10, we cannot initialize a MPIController with an external MPI_Comm */
    /* initialize the VTK communicator from the current MPI communicator */
    if(M_mpiComm)
    {
        M_opaqueComm = new vtkMPICommunicatorOpaqueComm(&M_mpiComm);
    }

    /* initialize in-situ visualization if needed */
    if(M_inSituProcessor == NULL)
    {
        M_inSituProcessor = vtkSmartPointer<vtkCPProcessor>::New();
        if(M_opaqueComm)
        {
            M_inSituProcessor->Initialize(*(M_opaqueComm));
        }
        //M_inSituProcessor->DebugOn();
    }
    else
    {
        M_inSituProcessor->RemoveAllPipelines();
    }

    /* specify a python user script */
    if(M_pythonPipeline != "")
    {
        std::cout << "Using Python script: " << M_pythonPipeline << std::endl;
        vtkSmartPointer<vtkCPPythonScriptPipeline> pipeline = vtkSmartPointer<vtkCPPythonScriptPipeline>::New();

        pipeline->Initialize(M_pythonPipeline.c_str());
        M_inSituProcessor->AddPipeline(pipeline.GetPointer());
    }
    /* else revert to a basic VTK pipeline */
    else
    {
        std::cout << "Using C++ pipeline" << std::endl;
        vtkSmartPointer<vtkBaseInsituPipeline> pipeline = vtkSmartPointer<vtkBaseInsituPipeline>::New();
        pipeline->Initialize(M_steeringEnabled);
        M_inSituProcessor->AddPipeline(pipeline.GetPointer());
    }

    M_pxm = vtkSMProxyManager::GetProxyManager();
#endif

    M_initialized = true;
}

void Catalyzer::GetSteeringValue( double * v, int * valid )
{
    /* set the value to invalid by default */
    if(valid)
    { *valid = 0; }

    /* First check if we use steering */
    if(M_steeringEnabled && M_pxm)
    {
        /* Then try to access the proxy */
        vtkSMProxy * px = M_pxm->GetProxy("sources", "SteeringPlugin1");
        if(px)
        {
            /* Get the client side object in the correct type */
            vtkObjectBase* clientSideObject = px->GetClientSideObject();
            vtkSteeringPlugin* realProducer = vtkSteeringPlugin::SafeDownCast(clientSideObject);

            /* set the value from the steering plugin */
            if(v)
            {
                *v = realProducer->GetValue();
                if(valid)
                { *valid = 1; }
            }
        }
    }
}

vtkSmartPointer<vtkMultiBlockDataSet>
Catalyzer::BuildMultiBlockDataSet( double time, vtkSmartPointer<vtkDataSet> out )
{
    std::ostringstream oss;

#if VTK_MAJOR_VERSION >= 6 && defined(VTK_HAS_PARALLEL)
    vtkSmartPointer<vtkMultiBlockDataSet> mbds = vtkSmartPointer<vtkMultiBlockDataSet>::New();

    int worldSize = 1; 
    int worldRank = 0; 
    if(M_mpiComm)
    {
        MPI_Comm_size(M_mpiComm, &worldSize); 
        MPI_Comm_rank(M_mpiComm, &worldRank); 
    }

    mbds->SetNumberOfBlocks( worldSize );
    
    /* Set the block corresponding to the processor on which we are working on */
    for( unsigned int block = 0 ; block < mbds->GetNumberOfBlocks(); ++block )
    {
        oss.str("");
        oss << "P" << block;
        /* If we own the block */
        if( block == worldRank )
        {
            mbds->SetBlock( block, out );
        }
        /* if we don't own the block set it to NULL */
        else
        {
            mbds->SetBlock( block, NULL );
        }
        mbds->GetMetaData(block)->Set(vtkCompositeDataSet::NAME(), oss.str().c_str() );
        mbds->GetMetaData(block)->Set(vtkDataObject::DATA_TIME_STEP(), time);
    }
    
#endif

    return mbds;
}

void
Catalyzer::Write( int stepIndex, std::string filename, vtkSmartPointer<vtkMultiBlockDataSet> mbds )
{
    /* InitializeExternal is only supported from 5.10+, */
    /* but lets aim for the latest major version 6 to reduce the complexity */
#if VTK_MAJOR_VERSION >= 6 && defined(VTK_HAS_PARALLEL)
    int worldSize = 1; 
    int worldRank = 0; 
    if(M_mpiComm)
    {
        MPI_Comm_size(M_mpiComm, &worldSize); 
        MPI_Comm_rank(M_mpiComm, &worldRank); 
    }

    /* Build vtk objects while reusing the current mpi communicator */
    /* before version 5.10, we cannot initialize a MPIController with an external MPI_Comm */
    vtkSmartPointer<vtkMPICommunicator> mpicomm = vtkSmartPointer<vtkMPICommunicator>::New();
    if(M_opaqueComm)
    {
        mpicomm->InitializeExternal(M_opaqueComm);
    }
    vtkSmartPointer<vtkMPIController> mpictrl = vtkSmartPointer<vtkMPIController>::New();
        mpictrl->SetCommunicator(mpicomm);

    vtkSmartPointer<vtkXMLPMultiBlockDataWriter> xmlpw = vtkSmartPointer<vtkXMLPMultiBlockDataWriter>::New();
        xmlpw->SetController(mpictrl);
        xmlpw->SetTimeStep(stepIndex);
        xmlpw->SetFileName(filename.c_str());
        xmlpw->SetInputData(mbds);
        /* only write the meta file on the first processor */
        if( worldRank == 0 )
        { xmlpw->SetWriteMetaFile(1); }
        else
        { xmlpw->SetWriteMetaFile(0); }
        xmlpw->Update();
#else
    vtkSmartPointer<vtkXMLMultiBlockDataWriter> mbw = vtkSmartPointer<vtkXMLMultiBlockDataWriter>::New();
        mbw->SetTimeStep(stepIndex);
        mbw->SetFileName(filename.c_str());
#if VTK_MAJOR_VERSION <= 5
        mbw->SetInput(mbds);
#else
        mbw->SetInputData(mbds);
#endif
        mbw->Update();
#endif
}

void
Catalyzer::Save(int index, double time, vtkSmartPointer<vtkDataSet> out)
{
    int stepIndex = index;
    double stepTime = time;
    std::ostringstream fname;
    std::ostringstream oss;

    /* The user must have forgotten to call init() */
    if(!M_initialized)
    {
        this->Initialize();
    }

    /* Build a multi block dataset based on gathered data */
    vtkSmartPointer<vtkMultiBlockDataSet> mbds = this->BuildMultiBlockDataSet( stepTime, out );

    /* initialize in-situ visualization if needed and if we specified pipelines to handle */
    if(M_inSituProcessor->GetNumberOfPipelines() > 0)
    {
        std::cout << "Processing timestep In-Situ " << stepIndex << " " << stepTime << std::endl;
        vtkSmartPointer<vtkCPDataDescription> dataDescription = vtkSmartPointer<vtkCPDataDescription>::New();
        dataDescription->AddInput("input");
        dataDescription->SetTimeData(time, stepIndex);

        vtkSmartPointer<vtkCharArray> hname = vtkSmartPointer<vtkCharArray>::New();
        hname->SetName("hostname");
        hname->SetNumberOfComponents(M_insituHostname.size() + 1);
        hname->InsertNextTupleValue(M_insituHostname.c_str());

        vtkSmartPointer<vtkIntArray> port = vtkSmartPointer<vtkIntArray>::New();
        port->SetName("port");
        port->InsertNextValue( M_insituPort );

        vtkSmartPointer<vtkFieldData> fdata = vtkSmartPointer<vtkFieldData>::New();
        fdata->AddArray(hname);
        fdata->AddArray(port);

        dataDescription->SetUserData(fdata);

        if(M_inSituProcessor->RequestDataDescription(dataDescription.GetPointer()) != 0)
        {
            dataDescription->GetInputDescriptionByName("input")->SetGrid(mbds);
            dataDescription->SetForceOutput(true);
            std::cout << "CoProcess " << M_inSituProcessor->CoProcess(dataDescription.GetPointer()) << std::endl;
        }
    }

    /* if insitu is not enable, or if it is enabled but we want to save data */
    /* handle thoses cases with this if */
#if 0
    if(M_inSituProcessor->GetNumberOfPipelines() == 0)
    {
        fname.str("");
        fname << "./data-" << (stepIndex);
#if VTK_MAJOR_VERSION < 6 || !defined(VTK_HAS_PARALLEL)
        int worldSize = 1; 
        int worldRank = 0; 
        if(M_mpiComm)
        {
            MPI_Comm_size(M_mpiComm, &worldSize); 
            MPI_Comm_rank(M_mpiComm, &worldRank); 
        }

        fname << "-" << worldSize << "_" << worldRank;
#endif
        fname << ".vtm";

        /* write VTK files */
        this->Write(stepIndex, fname.str(), mbds);
    }
#endif
}

#endif // __EXPORTERVTK_CPP
