#include "hydroParams.h"

#include <stdlib.h> // for exit
#include <stdio.h>  // for fprintf
#include <string.h> // for strcmp

#include <mpi.h>

#include "ini/ini.h" // our INI file reader

const char * varNames[4] = { "rho", "E", "mx", "my" };

/*
 * Hydro Parameters (default values)
 */

// run parameters
int    nStepmax=0; /*!< maximun number of time steps. */
real_t tEnd=0.0;   /*!< end of simulation time. */
int    nOutput;    /*!< number of time steps between 2 consecutive outputs. */

// geometry parameters
int nx=0;     /*!< logical size along X (without ghost cells).*/
int ny=0;     /*!< logical size along Y (without ghost cells).*/
const int ghostWidth=2;  
int imin=0;   /*!< index minimum at X border*/
int imax=0;   /*!< index maximum at X border*/
int jmin=0;   /*!< index minimum at Y border*/
int jmax=0;   /*!< index maximum at Y border*/

int isize=0;  /*!< total size (in cell unit) along X direction with ghosts.*/
int jsize=0;  /*!< total size (in cell unit) along Y direction with ghosts.*/

real_t xmin=0.0;
real_t xmax=1.0;
real_t ymin=0.0;
real_t ymax=1.0;
real_t dx;       /*!< x resolution */
real_t dy;       /*!< y resolution */

int boundary_type_xmin=1;
int boundary_type_xmax=1;
int boundary_type_ymin=1;
int boundary_type_ymax=1;

// MPI parameters
int mx = 1;
int my = 1;
int iProc = 0;
int jProc = 0;

// IO parameters
bool ioVTK=true;    /*!< enable VTK  output file format (using VTI).*/
bool ioHDF5=false;  /*!< enable HDF5 output file format.*/

// hydro (numerical scheme) parameters
real_t gamma0;      /*!< specific heat capacity ratio (adiabatic index)*/
real_t gamma6;
real_t cfl;         /*!< Courant-Friedrich-Lewy parameter.*/
real_t slope_type;  /*!< type of slope computation (2 for second order scheme).*/
int    iorder;      /*!< */
real_t smallr=1e-8; /*!< small density cut-off*/
real_t smallc=1e-8; /*!< small speed of sound cut-off*/
real_t smallp;      /*!< small pressure cut-off*/
real_t smallpp;     /*!< smallp times smallr*/

int niter_riemann;  /*!< number of iteration usd in quasi-exact riemann solver*/
int riemannSolverType;
char problem[256] = "implode";

// blast problem parameters
real_t blast_radius=10.0;
int blast_center_x=0;
int blast_center_y=0;
real_t blast_density_in=1.0;
real_t blast_density_out=1.2;
real_t blast_pressure_in=10.0;
real_t blast_pressure_out=0.1;

// other parameters
Catalyzer * catalyzer=NULL;
//int nbVar=4;  /*!< number of fields in simulation (density, energy, vx, vy)*/
int implementationVersion=0; /*!< triggers which implementation to use (currently 2 versions)*/

// =======================================================
// =======================================================
/*
 * Handler routine used to parse value from ini file.
 */
static int handler(void* user, 
		   const char* section, 
		   const char* name,
                   const char* value)
{
  
#define MATCH(s, n) ( strcmp(section, s) == 0 && strcmp(name, n) == 0 )
  if (MATCH("RUN", "tEnd")) {
    tEnd = atof(value);
  } else if (MATCH("RUN", "nStepmax")) {
    nStepmax = atoi(value);
  } else if (MATCH("RUN", "nOutput")) {
    nOutput = atoi(value);
  } else if (MATCH("MESH", "nx")) {
    nx = atoi(value);
  } else if (MATCH("MESH", "ny")) {
    ny = atoi(value);
  } else if (MATCH("MESH", "mx")) {
    mx = atoi(value);
  } else if (MATCH("MESH", "my")) {
    my = atoi(value);
  } else if (MATCH("MESH", "boundary_type_xmin")) {
    boundary_type_xmin = atoi(value);
  } else if (MATCH("MESH", "boundary_type_xmax")) {
    boundary_type_xmax = atoi(value);
  } else if (MATCH("MESH", "boundary_type_ymin")) {
    boundary_type_ymin = atoi(value);
  } else if (MATCH("MESH", "boundary_type_ymax")) {
    boundary_type_ymax = atoi(value);
  } else if (MATCH("HYDRO", "gamma0")) {
    gamma0 = atof(value);
  } else if (MATCH("HYDRO", "cfl")) {
    cfl = atof(value);
  } else if (MATCH("HYDRO", "niter_riemann")) {
    niter_riemann = atoi(value);
  } else if (MATCH("HYDRO", "iorder")) {
    iorder = atoi(value);
  } else if (MATCH("HYDRO", "slope_type")) {
    slope_type = atoi(value);
  } else if (MATCH("HYDRO", "smallc")) {
    smallc = atof(value);
  } else if (MATCH("HYDRO", "smallr")) {
    smallr = atof(value);
  } else if (MATCH("HYDRO", "problem")) {
    strcpy(problem,value);
  } else if (MATCH("blast", "radius")) {
    blast_radius = atof(value);
  } else if (MATCH("blast", "center_x")) {
    blast_center_x = atoi(value);
  } else if (MATCH("blast", "center_y")) {
    blast_center_y = atoi(value);
  } else if (MATCH("blast", "density_in")) {
    blast_density_in = atof(value);
  } else if (MATCH("blast", "density_out")) {
    blast_density_out = atof(value);
  } else if (MATCH("blast", "pressure_in")) {
    blast_pressure_in = atof(value);
  } else if (MATCH("blast", "pressure_out")) {
    blast_pressure_out = atof(value);
  } else if (MATCH("OTHER", "implementationVersion")) {
    implementationVersion = atoi(value);
  } else {
    return 0;  /* unknown section/name, error */
  }
  return 1;
}

// =======================================================
// =======================================================
void initHydroParameters(int argc, char* argv[])
{

  /* MPI parameters */
  int  myRank, nProc;
  MPI_Comm_size(MPI_COMM_WORLD, &nProc);
  MPI_Comm_rank(MPI_COMM_WORLD, &myRank);

  // assumes argc == 2, input file name given as the only parameter on the command line
  if (argc != 2) {
    int i;
    fprintf(stderr, "Error: wrong number of argument (%d); input filename must be the only parameter on the command line\n",argc);
    for (i=0; i<argc; i++) {
      fprintf(stderr,"argv[%d] = %s\n",i,argv[i]);
    }
    exit(EXIT_FAILURE);
  }

  // parse ini file
  if (ini_parse(argv[1], handler, NULL) < 0) {
    fprintf(stderr,"Can't load 'test.ini'\n");
    exit(EXIT_FAILURE);
  }

  // set other parameters
  imax = nx-1+2*ghostWidth;
  jmax = ny-1+2*ghostWidth;
  
  isize = imax - imin + 1;
  jsize = jmax - jmin + 1;
  
  // take care that dx must use global resolution
  dx = (xmax - xmin) / (nx*mx);
  dy = (ymax - ymin) / (ny*my);

  // MPI grid : myRank = iProc + mx*jProc
  jProc = myRank/mx;
  iProc = myRank-mx*jProc;
  
  smallp  = smallc*smallc/gamma0;
  smallpp = smallr*smallp;
  gamma6  = (gamma0 + ONE_F)/(TWO_F * gamma0);
  
  // check that given parameters are valid
  if ( (implementationVersion != 0) && (implementationVersion != 1) ) {
    fprintf(stderr, "The implementation version parameter should 0 or 1 !!!");
    fprintf(stderr, "Check your parameter file, section OTHER");
    exit(EXIT_FAILURE);
  } else {
    fprintf(stdout, "Using implementation version %d\n", implementationVersion);
  }

} // initHydroParameters


// =======================================================
// =======================================================
void printHydroParameters()
{
  printf( "##########################\n");
  printf( "Simulation run parameters:\n");
  printf( "##########################\n");
  printf( "nx         : %d\n", nx);
  printf( "ny         : %d\n", ny);
  printf( "mx         : %d\n", mx);
  printf( "my         : %d\n", my);
  printf( "Global size x : %d\n", nx*mx);
  printf( "Global size y : %d\n", ny*my);
  printf( "dx         : %f\n", dx);
  printf( "dy         : %f\n", dy);
  printf( "imin       : %d\n", imin);
  printf( "imax       : %d\n", imax);
  printf( "jmin       : %d\n", jmin);      
  printf( "jmax       : %d\n", jmax);      
  printf( "nStepmax   : %d\n",nStepmax);
  printf( "tEnd       : %f\n",tEnd);
  printf( "nOutput    : %d\n",nOutput);
  printf( "gamma0     : %f\n",gamma0);
  printf( "cfl        : %f\n",cfl);
  printf( "smallr     : %12.10f\n",smallr);
  printf( "smallc     : %12.10f\n",smallc);
  //printf( "niter_riemann : %d\n",niter_riemann);
  printf( "iorder     : %d\n",iorder);
  printf( "slope_type : %f\n",slope_type);
  printf( "problem    : %s\n",problem);
  printf( "implementation version : %d\n",implementationVersion);
  printf( "##########################\n");
  
} // printHydroParameters
