#include <stdio.h>
#include <math.h>

#include "hydroUtils.h"

#include "hydroParams.h" // for gamma0, smallc, etc ...

#define SATURATE saturate_cpu
static real_t saturate_cpu(real_t a) {
  
  if (isnan(a)) return ZERO_F;
  return a >= ONE_F ? ONE_F : a <= ZERO_F ? ZERO_F : a;

} // saturate_cpu

void swapValues(real_t *a, real_t *b) {

  real_t tmp = *a;

  *a = *b;
  *b = tmp;

} // swapValues

/**
 * Equation of state:
 * compute pressure p and speed of sound c, from density rho and
 * internal energy eint using the "calorically perfect gas" equation
 * of state : \f$ eint=\frac{p}{\rho (\gamma-1)} \f$
 * Recall that \f$ \gamma \f$ is equal to the ratio of specific heats \f$ \left[
 * c_p/c_v \right] \f$.
 * 
 * @param[in]  rho  density
 * @param[in]  eint internal energy
 * @param[out] p    pressure
 * @param[out] c    speed of sound
 */
void eos(real_t rho, real_t eint, real_t* p, real_t* c)
{
  *p = FMAX((gamma0 - ONE_F) * rho * eint, rho * smallp);
  *c = SQRT(gamma0 * (*p) / rho);
} // eos

/**
 * Convert conservative variables (rho, rho*u, rho*v, e) to 
 * primitive variables (rho,u,v,p)
 * @param[in]  u  conservative variables array
 * @param[out] q  primitive    variables array (allocated in calling routine, size is constant nbVar)
 * @param[out] c  local speed of sound
 */
void computePrimitives(const real_t u[nbVar], real_t* c, real_t q[nbVar])
{
  q[ID] = FMAX(u[ID], smallr);
  q[IU] = u[IU] / q[ID];
  q[IV] = u[IV] / q[ID];
  
  real_t eken = HALF_F * (q[IU] * q[IU] + q[IV] * q[IV]);
  real_t e = u[IP] / q[ID] - eken;
  if (e < 0) {
    printf("FATAL ERROR : hydro eint < 0  : e %f eken %f d %f u %f v %f\n",u[IP],eken,u[ID],u[IU],u[IV]);
    exit(0);
  }
  
  // compute pressure and speed of sound
  eos(q[ID], e, &(q[IP]), c);

} // computePrimitive

/**
 * Convert conservative variables (rho, rho*u, rho*v, e) to 
 * primitive variables (rho,u,v,p)
 * @param[in]  U    conservative variables (2D array)
 * @param[in]  i    x-coordinate of current cell
 * @param[in]  j    y-coordinate of current cell
 * @param[out] qLoc primitive variables of current cell (allocated in calling routine, size is constant nbVar)
 * @param[out] c  local speed of sound
 */
void computePrimitives_ij(const HydroArray2d* U, int i, int j, real_t* c, real_t qLoc[nbVar])
{

  real_t uLoc[nbVar];

  // get primitive variables in current cell
  uLoc[ID] = HydroArray2d_get(U,i,j,ID);
  uLoc[IP] = HydroArray2d_get(U,i,j,IP);
  uLoc[IU] = HydroArray2d_get(U,i,j,IU);
  uLoc[IV] = HydroArray2d_get(U,i,j,IV);

  computePrimitives(uLoc, c, qLoc);

  /* if (i==2 && j==2) { */
  /*   printf("compute_dt %f %f %f %f\n",uLoc[ID], uLoc[IP], uLoc[IU], uLoc[IV]); */
  /*   printf("compute_dt # %f %f %f %f\n",qLoc[ID], qLoc[IP], qLoc[IU], qLoc[IV]); */
  /* } */

} // computePrimitive_ij

/**
 * Trace computations for unsplit Godunov scheme.
 *
 * \param[in] q          : Primitive variables state.
 * \param[in] qNeighbors : state in the neighbor cells (2 neighbors
 * per dimension, in the following order x+, x-, y+, y-, z+, z-)
 * \param[in] c          : local sound speed.
 * \param[in] dtdx       : dt over dx
 * \param[out] qm        : qm state (one per dimension)
 * \param[out] qp        : qp state (one per dimension)
 */
void trace_unsplit_2d(const real_t q[nbVar], 
		      const real_t qNeighbors[2*NDIM][nbVar],
		      real_t c, 
		      real_t dtdx, 
		      real_t dtdy, 
		      real_t qm[NDIM][nbVar], 
		      real_t qp[NDIM][nbVar])
{
  // first compute slopes
  real_t dqX[nbVar];
  real_t dqY[nbVar];
  
  slope_unsplit_hydro_2d(q, 
			 qNeighbors[0], qNeighbors[1], 
			 qNeighbors[2], qNeighbors[3],
			 dqX, dqY);
  
  // Cell centered values
  real_t r =  q[ID];
  real_t p =  q[IP];
  real_t u =  q[IU];
  real_t v =  q[IV];
  
  // TVD slopes in all directions
  real_t drx = dqX[ID];
  real_t dpx = dqX[IP];
  real_t dux = dqX[IU];
  real_t dvx = dqX[IV];
  
  real_t dry = dqY[ID];
  real_t dpy = dqY[IP];
  real_t duy = dqY[IU];
  real_t dvy = dqY[IV];
  
  // source terms (with transverse derivatives)
  real_t sr0 = -u*drx-v*dry - (dux+dvy)*r;
  real_t sp0 = -u*dpx-v*dpy - (dux+dvy)*gamma0*p;
  real_t su0 = -u*dux-v*duy - (dpx    )/r;
  real_t sv0 = -u*dvx-v*dvy - (dpy    )/r;
  
  // Right state at left interface
  qp[0][ID] = r - HALF_F*drx + sr0*dtdx*HALF_F;
  qp[0][IP] = p - HALF_F*dpx + sp0*dtdx*HALF_F;
  qp[0][IU] = u - HALF_F*dux + su0*dtdx*HALF_F;
  qp[0][IV] = v - HALF_F*dvx + sv0*dtdx*HALF_F;
  qp[0][ID] = FMAX(smallr, qp[0][ID]);
  
  // Left state at right interface
  qm[0][ID] = r + HALF_F*drx + sr0*dtdx*HALF_F;
  qm[0][IP] = p + HALF_F*dpx + sp0*dtdx*HALF_F;
  qm[0][IU] = u + HALF_F*dux + su0*dtdx*HALF_F;
  qm[0][IV] = v + HALF_F*dvx + sv0*dtdx*HALF_F;
  qm[0][ID] = FMAX(smallr, qm[0][ID]);
  
  // Top state at bottom interface
  qp[1][ID] = r - HALF_F*dry + sr0*dtdy*HALF_F;
  qp[1][IP] = p - HALF_F*dpy + sp0*dtdy*HALF_F;
  qp[1][IU] = u - HALF_F*duy + su0*dtdy*HALF_F;
  qp[1][IV] = v - HALF_F*dvy + sv0*dtdy*HALF_F;
  qp[1][ID] = FMAX(smallr, qp[1][ID]);
  
  // Bottom state at top interface
  qm[1][ID] = r + HALF_F*dry + sr0*dtdy*HALF_F;
  qm[1][IP] = p + HALF_F*dpy + sp0*dtdy*HALF_F;
  qm[1][IU] = u + HALF_F*duy + su0*dtdy*HALF_F;
  qm[1][IV] = v + HALF_F*dvy + sv0*dtdy*HALF_F;
  qm[1][ID] = FMAX(smallr, qm[1][ID]);
  
} // trace_unsplit_2d

/**
 * Trace computations for unsplit Godunov scheme.
 *
 * \param[in] q          : Primitive variables state.
 * \param[in] dqX        : slope along X
 * \param[in] dqY        : slope along Y
 * \param[in] c          : local sound speed.
 * \param[in] dtdx       : dt over dx
 * \param[in] dtdy       : dt over dy
 * \param[in] faceId     : which face will be reconstructed
 * \param[out] qface     : q reconstructed state at cell interface
 */
void trace_unsplit_2d_along_dir(const real_t q[nbVar], 
				const real_t dqX[nbVar],
				const real_t dqY[nbVar],
				real_t dtdx, 
				real_t dtdy, 
				int    faceId,
				real_t qface[nbVar])
{
  
  // Cell centered values
  real_t r =  q[ID];
  real_t p =  q[IP];
  real_t u =  q[IU];
  real_t v =  q[IV];
  
  // TVD slopes in all directions
  real_t drx = dqX[ID];
  real_t dpx = dqX[IP];
  real_t dux = dqX[IU];
  real_t dvx = dqX[IV];
  
  real_t dry = dqY[ID];
  real_t dpy = dqY[IP];
  real_t duy = dqY[IU];
  real_t dvy = dqY[IV];
  
  // source terms (with transverse derivatives)
  real_t sr0 = -u*drx-v*dry - (dux+dvy)*r;
  real_t sp0 = -u*dpx-v*dpy - (dux+dvy)*gamma0*p;
  real_t su0 = -u*dux-v*duy - (dpx    )/r;
  real_t sv0 = -u*dvx-v*dvy - (dpy    )/r;
  
  if (faceId == FACE_XMIN) {
    // Right state at left interface
    qface[ID] = r - HALF_F*drx + sr0*dtdx*HALF_F;
    qface[IP] = p - HALF_F*dpx + sp0*dtdx*HALF_F;
    qface[IU] = u - HALF_F*dux + su0*dtdx*HALF_F;
    qface[IV] = v - HALF_F*dvx + sv0*dtdx*HALF_F;
    qface[ID] = FMAX(smallr, qface[ID]);
  }

  if (faceId == FACE_XMAX) {
    // Left state at right interface
    qface[ID] = r + HALF_F*drx + sr0*dtdx*HALF_F;
    qface[IP] = p + HALF_F*dpx + sp0*dtdx*HALF_F;
    qface[IU] = u + HALF_F*dux + su0*dtdx*HALF_F;
    qface[IV] = v + HALF_F*dvx + sv0*dtdx*HALF_F;
    qface[ID] = FMAX(smallr, qface[ID]);
  }
  
  if (faceId == FACE_YMIN) {
    // Top state at bottom interface
    qface[ID] = r - HALF_F*dry + sr0*dtdy*HALF_F;
    qface[IP] = p - HALF_F*dpy + sp0*dtdy*HALF_F;
    qface[IU] = u - HALF_F*duy + su0*dtdy*HALF_F;
    qface[IV] = v - HALF_F*dvy + sv0*dtdy*HALF_F;
    qface[ID] = FMAX(smallr, qface[ID]);
  }

  if (faceId == FACE_YMAX) {
    // Bottom state at top interface
    qface[ID] = r + HALF_F*dry + sr0*dtdy*HALF_F;
    qface[IP] = p + HALF_F*dpy + sp0*dtdy*HALF_F;
    qface[IU] = u + HALF_F*duy + su0*dtdy*HALF_F;
    qface[IV] = v + HALF_F*dvy + sv0*dtdy*HALF_F;
    qface[ID] = FMAX(smallr, qface[ID]);
  }

} // trace_unsplit_2d_along_dir

/**
 * This another implementation of trace computations for 2D data; it
 * is used when unsplitVersion = 1
 *
 * Note that :
 * - hydro slopes computations are done outside this routine
 *
 * \param[in]  q  primitive variable state vector
 * \param[in]  dq primitive variable slopes
 * \param[in]  dtdx dt divided by dx
 * \param[in]  dtdy dt divided by dy
 * \param[out] qm
 * \param[out] qp
 *
 */
void trace_unsplit_hydro_2d(const real_t q[nbVar],
			    const real_t dq[2][nbVar],
			    real_t dtdx,
			    real_t dtdy,
			    real_t qm[2][nbVar], 
			    real_t qp[2][nbVar])
{
  
  // Cell centered values
  real_t r = q[ID];
  real_t p = q[IP];
  real_t u = q[IU];
  real_t v = q[IV];

  // Cell centered TVD slopes in X direction
  real_t drx = dq[IX][ID];  drx *= HALF_F;
  real_t dpx = dq[IX][IP];  dpx *= HALF_F;
  real_t dux = dq[IX][IU];  dux *= HALF_F;
  real_t dvx = dq[IX][IV];  dvx *= HALF_F;
  
  // Cell centered TVD slopes in Y direction
  real_t dry = dq[IY][ID];  dry *= HALF_F;
  real_t dpy = dq[IY][IP];  dpy *= HALF_F;
  real_t duy = dq[IY][IU];  duy *= HALF_F;
  real_t dvy = dq[IY][IV];  dvy *= HALF_F;

  // Source terms (including transverse derivatives)
  real_t sr0, su0, sv0, sp0;

  /*only true for cartesian grid */
  {
    sr0 = (-u*drx-dux*r)       *dtdx + (-v*dry-dvy*r)       *dtdy;
    su0 = (-u*dux-dpx/r)       *dtdx + (-v*duy      )       *dtdy;
    sv0 = (-u*dvx      )       *dtdx + (-v*dvy-dpy/r)       *dtdy;
    sp0 = (-u*dpx-dux*gamma0*p)*dtdx + (-v*dpy-dvy*gamma0*p)*dtdy;    
  } // end cartesian

  // Update in time the  primitive variables
  r = r + sr0;
  u = u + su0;
  v = v + sv0;
  p = p + sp0;

  // Face averaged right state at left interface
  qp[0][ID] = r - drx;
  qp[0][IU] = u - dux;
  qp[0][IV] = v - dvx;
  qp[0][IP] = p - dpx;
  qp[0][ID] = FMAX(smallr,  qp[0][ID]);
  qp[0][IP] = FMAX(smallp * qp[0][ID], qp[0][IP]);
  
  // Face averaged left state at right interface
  qm[0][ID] = r + drx;
  qm[0][IU] = u + dux;
  qm[0][IV] = v + dvx;
  qm[0][IP] = p + dpx;
  qm[0][ID] = FMAX(smallr,  qm[0][ID]);
  qm[0][IP] = FMAX(smallp * qm[0][ID], qm[0][IP]);

  // Face averaged top state at bottom interface
  qp[1][ID] = r - dry;
  qp[1][IU] = u - duy;
  qp[1][IV] = v - dvy;
  qp[1][IP] = p - dpy;
  qp[1][ID] = FMAX(smallr,  qp[1][ID]);
  qp[1][IP] = FMAX(smallp * qp[1][ID], qp[1][IP]);
  
  // Face averaged bottom state at top interface
  qm[1][ID] = r + dry;
  qm[1][IU] = u + duy;
  qm[1][IV] = v + dvy;
  qm[1][IP] = p + dpy;
  qm[1][ID] = FMAX(smallr,  qm[1][ID]);
  qm[1][IP] = FMAX(smallp * qm[1][ID], qm[1][IP]);
  
} // trace_unsplit_hydro_2d

/**
 * Compute primitive variables slope (vector dq) from q and its neighbors.
 * This routine is only used in the 2D UNSPLIT integration and slope_type = 0,1 and 2.
 * 
 * Only slope_type 1 and 2 are supported.
 *
 * \param[in]  q       : current primitive variable state
 * \param[in]  qPlusX  : state in the next neighbor cell along XDIR
 * \param[in]  qMinusX : state in the previous neighbor cell along XDIR
 * \param[in]  qPlusY  : state in the next neighbor cell along YDIR
 * \param[in]  qMinusY : state in the previous neighbor cell along YDIR
 * \param[out] dqX     : reference to an array returning the X slopes
 * \param[out] dqY     : reference to an array returning the Y slopes
 *
 */
void slope_unsplit_hydro_2d(const real_t q[nbVar], 
			    const real_t qPlusX[nbVar], 
			    const real_t qMinusX[nbVar],
			    const real_t qPlusY[nbVar], 
			    const real_t qMinusY[nbVar],
			    real_t dqX[nbVar],
			    real_t dqY[nbVar])
{
  
  if (slope_type==0) {
    for (int nVar=0; nVar<nbVar; ++nVar) {
      dqX[nVar] = ZERO_F;
      dqY[nVar] = ZERO_F;
    }
    return;
  }

  if (slope_type==1 || slope_type==2) {  // minmod or average

    real_t dlft, drgt, dcen, dsgn, slop, dlim;
    
    for (int nVar=0; nVar<nbVar; ++nVar) {

      // slopes in first coordinate direction
      dlft = slope_type*(q     [nVar] - qMinusX[nVar]);
      drgt = slope_type*(qPlusX[nVar] - q      [nVar]);
      dcen = HALF_F * (qPlusX[nVar] - qMinusX[nVar]);
      dsgn = (dcen >= ZERO_F) ? ONE_F : -ONE_F;
      slop = FMIN( FABS(dlft), FABS(drgt) );
      dlim = slop;
      if ( (dlft*drgt) <= ZERO_F )
	dlim = ZERO_F;
      dqX[nVar] = dsgn * FMIN( dlim, FABS(dcen) );
      
      // slopes in second coordinate direction
      dlft = slope_type*(q     [nVar] - qMinusY[nVar]);
      drgt = slope_type*(qPlusY[nVar] - q      [nVar]);
      dcen = HALF_F * (qPlusY[nVar] - qMinusY[nVar]);
      dsgn = (dcen >= ZERO_F) ? ONE_F : -ONE_F;
      slop = FMIN( FABS(dlft), FABS(drgt) );
      dlim = slop;
      if ( (dlft*drgt) <= ZERO_F )
	dlim = ZERO_F;
      dqY[nVar] = dsgn * FMIN( dlim, FABS(dcen) );

    } // end for nVar

  } // end slope_type == 1 or 2
  
} // slope_unsplit_hydro_2d

/**
 * Compute cell fluxes from the Godunov state
 * \param[in]  qgdnv input Godunov state
 * \param[out] flux  output flux vector
 */
void cmpflx(const real_t qgdnv[nbVar], 
	    real_t flux[nbVar])
{
  // Compute fluxes
  // Mass density
  flux[ID] = qgdnv[ID] * qgdnv[IU];
  
  // Normal momentum
  flux[IU] = flux[ID] * qgdnv[IU] + qgdnv[IP];
  
  // Transverse momentum
  flux[IV] = flux[ID] * qgdnv[IV];

  // Total energy
  real_t entho = ONE_F / (gamma0 - ONE_F);
  real_t ekin;
  ekin = HALF_F * qgdnv[ID] * (qgdnv[IU]*qgdnv[IU] + qgdnv[IV]*qgdnv[IV]);
  
  real_t etot = qgdnv[IP] * entho + ekin;
  flux[IP] = qgdnv[IU] * (etot + qgdnv[IP]);

} // cmpflx

/** 
 * Riemann solver, equivalent to riemann_approx in RAMSES (see file
 * godunov_utils.f90 in RAMSES).
 * 
 * @param[in] qleft  : input left state
 * @param[in] qright : input right state
 * @param[out] qgdnv : output Godunov state
 * @param[out] flux  : output flux
 */
void riemann_approx(const real_t qleft[nbVar], 
		    const real_t qright[nbVar], 
		    real_t qgdnv[nbVar], 
		    real_t flux[nbVar])
{

  // Pressure, density and velocity
  real_t rl = FMAX(qleft [ID], smallr);
  real_t ul =      qleft [IU];
  real_t pl = FMAX(qleft [IP], rl*smallp);
  real_t rr = FMAX(qright[ID], smallr);
  real_t ur =      qright[IU];
  real_t pr = FMAX(qright[IP], rr*smallp);
  
  // Lagrangian sound speed
  real_t cl = gamma0*pl*rl;
  real_t cr = gamma0*pr*rr;
  
  // First guess
  real_t wl = SQRT(cl);
  real_t wr = SQRT(cr);
  real_t pstar = FMAX(((wr*pl+wl*pr)+wl*wr*(ul-ur))/(wl+wr), (real_t) ZERO_F);
  real_t pold = pstar;
  real_t conv = ONE_F;
  
  // Newton-Raphson iterations to find pstar at the required accuracy
  for(int iter = 0; (iter < niter_riemann) && (conv > 1e-6); ++iter)
    {
      real_t wwl = SQRT(cl*(ONE_F+gamma6*(pold-pl)/pl));
      real_t wwr = SQRT(cr*(ONE_F+gamma6*(pold-pr)/pr));
      real_t ql = 2.0f*wwl*wwl*wwl/(wwl*wwl+cl);
      real_t qr = 2.0f*wwr*wwr*wwr/(wwr*wwr+cr);
      real_t usl = ul-(pold-pl)/wwl;
      real_t usr = ur+(pold-pr)/wwr;
      real_t delp = FMAX(qr*ql/(qr+ql)*(usl-usr),-pold);
      
      pold = pold+delp;
      conv = FABS(delp/(pold+smallpp));	 // Convergence indicator
    }
  
  // Star region pressure
  // for a two-shock Riemann problem
  pstar = pold;
  wl = SQRT(cl*(ONE_F+gamma6*(pstar-pl)/pl));
  wr = SQRT(cr*(ONE_F+gamma6*(pstar-pr)/pr));
  
  // Star region velocity
  // for a two shock Riemann problem
  real_t ustar = HALF_F * (ul + (pl-pstar)/wl + ur - (pr-pstar)/wr);
  
  // Left going or right going contact wave
  real_t sgnm = COPYSIGN(ONE_F, ustar);
  
  // Left or right unperturbed state
  real_t ro, uo, po, wo;
  if(sgnm > ZERO_F)
    {
      ro = rl;
      uo = ul;
      po = pl;
      wo = wl;
    }
  else
    {
      ro = rr;
      uo = ur;
      po = pr;
      wo = wr;
    }
  real_t co = FMAX(smallc, SQRT(FABS(gamma0*po/ro)));
  
  // Star region density (Shock, FMAX prevents vacuum formation in star region)
  real_t rstar = FMAX((real_t) (ro/(ONE_F+ro*(po-pstar)/(wo*wo))), (real_t) (smallr));
  // Star region sound speed
  real_t cstar = FMAX(smallc, SQRT(FABS(gamma0*pstar/rstar)));
  
  // Compute rarefaction head and tail speed
  real_t spout  = co    - sgnm*uo;
  real_t spin   = cstar - sgnm*ustar;
  // Compute shock speed
  real_t ushock = wo/ro - sgnm*uo;
  
  if(pstar >= po)
    {
      spin  = ushock;
      spout = ushock;
    }
  
  // Sample the solution at x/t=0
  real_t scr = FMAX(spout-spin, smallc+FABS(spout+spin));
  real_t frac = HALF_F * (ONE_F + (spout + spin)/scr);
  //real_t frac = SATURATE();
  if (isnan(frac))
    frac = ZERO_F;
  else
    frac = SATURATE(frac);
  
  qgdnv[ID] = frac*rstar + (ONE_F-frac)*ro;
  qgdnv[IU] = frac*ustar + (ONE_F-frac)*uo;
  qgdnv[IP] = frac*pstar + (ONE_F-frac)*po;
  
  if(spout < ZERO_F)
    {
      qgdnv[ID] = ro;
      qgdnv[IU] = uo;
      qgdnv[IP] = po;
    }
  
  if(spin > ZERO_F)
    {
      qgdnv[ID] = rstar;
      qgdnv[IU] = ustar;
      qgdnv[IP] = pstar;
    }
  
  // transverse velocity
  if(sgnm > ZERO_F)
    {
      qgdnv[IV] = qleft[IV];
    }
  else
    {
      qgdnv[IV] = qright[IV];
    }
  
  cmpflx(qgdnv, flux);
  
} // riemann_approx

/** 
 * Riemann solver HLLC
 *
 * @param[in] qleft : input left state
 * @param[in] qright : input right state
 * @param[out] qgdnv : output Godunov state
 * @param[out] flux  : output flux
 */
void riemann_hllc(const real_t qleft[nbVar], 
		  const real_t qright[nbVar], 
		  real_t qgdnv[nbVar], 
		  real_t flux[nbVar])
{

  const real_t entho = ONE_F / (gamma0 - ONE_F);
  
  // Left variables
  real_t rl = FMAX(qleft[ID], smallr);
  real_t pl = FMAX(qleft[IP], rl*smallp);
  real_t ul =      qleft[IU];
    
  real_t ecinl = HALF_F*rl*ul*ul;
  ecinl += HALF_F*rl*qleft[IV]*qleft[IV];

  real_t etotl = pl*entho+ecinl;
  real_t ptotl = pl;

  // Right variables
  real_t rr = FMAX(qright[ID], smallr);
  real_t pr = FMAX(qright[IP], rr*smallp);
  real_t ur =      qright[IU];

  real_t ecinr = HALF_F*rr*ur*ur;
  ecinr += HALF_F*rr*qright[IV]*qright[IV];
  
  real_t etotr = pr*entho+ecinr;
  real_t ptotr = pr;
    
  // Find the largest eigenvalues in the normal direction to the interface
  real_t cfastl = SQRT(FMAX(gamma0*pl/rl,smallc*smallc));
  real_t cfastr = SQRT(FMAX(gamma0*pr/rr,smallc*smallc));

  // Compute HLL wave speed
  real_t SL = FMIN(ul,ur) - FMAX(cfastl,cfastr);
  real_t SR = FMAX(ul,ur) + FMAX(cfastl,cfastr);

  // Compute lagrangian sound speed
  real_t rcl = rl*(ul-SL);
  real_t rcr = rr*(SR-ur);
    
  // Compute acoustic star state
  real_t ustar    = (rcr*ur   +rcl*ul   +  (ptotl-ptotr))/(rcr+rcl);
  real_t ptotstar = (rcr*ptotl+rcl*ptotr+rcl*rcr*(ul-ur))/(rcr+rcl);

  // Left star region variables
  real_t rstarl    = rl*(SL-ul)/(SL-ustar);
  real_t etotstarl = ((SL-ul)*etotl-ptotl*ul+ptotstar*ustar)/(SL-ustar);
    
  // Right star region variables
  real_t rstarr    = rr*(SR-ur)/(SR-ustar);
  real_t etotstarr = ((SR-ur)*etotr-ptotr*ur+ptotstar*ustar)/(SR-ustar);
    
  // Sample the solution at x/t=0
  real_t ro, uo, ptoto, etoto;
  if (SL > ZERO_F) {
    ro=rl;
    uo=ul;
    ptoto=ptotl;
    etoto=etotl;
  } else if (ustar > ZERO_F) {
    ro=rstarl;
    uo=ustar;
    ptoto=ptotstar;
    etoto=etotstarl;
  } else if (SR > ZERO_F) {
    ro=rstarr;
    uo=ustar;
    ptoto=ptotstar;
    etoto=etotstarr;
  } else {
    ro=rr;
    uo=ur;
    ptoto=ptotr;
    etoto=etotr;
  }
      
  // Compute the Godunov flux
  flux[ID] = ro*uo;
  flux[IU] = ro*uo*uo+ptoto;
  flux[IP] = (etoto+ptoto)*uo;
  if (flux[ID] > ZERO_F) {
    flux[IV] = flux[ID]*qleft[IV];
  } else {
    flux[IV] = flux[ID]*qright[IV];
  }
  
} // riemann_hllc
