/**
 * Hydro2d solver for MPI teaching purpose.
 *
 * \date July, 7 2013
 * \author P. Kestener
 */

#include <stdlib.h>
#include <stdio.h>

#include <mpi.h>

#include "real_type.h"   // choose between single and double precision
#include "hydroParams.h" // read parameter file
#include "hydroRun.h"    // memory allocation for hydro arrays
#include "monitoring.h"  // for timer

#include <catalyzer.h>

int main(int argc, char *argv[])
{

  /* MPI parameters */
  int  myRank, nProc;
  char processor_name[MPI_MAX_PROCESSOR_NAME];
  int  processor_name_len;

  /* simulation run parameters */
  real_t t=0, dt=0;
  int    nStep=0;
  
  HydroWorkspace workspace;

  // init MPI
  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &nProc);
  MPI_Comm_rank(MPI_COMM_WORLD, &myRank);
  MPI_Get_processor_name(processor_name, &processor_name_len);

  if (0==myRank) {
    printf("#################################################################\n");
    printf("Number of MPI processes=%03d | My_rank=%03d | processor_name=%5s\n",
	   nProc, myRank, processor_name);
    printf("#################################################################\n");
  }

  // read parameter file and initialize parameter
  initHydroParameters(argc, argv);

  // print parameters on screen
  if (myRank==0)
    printHydroParameters();

  // sanity check : MPI parameters valid ?
  if ( (mx*my) != nProc ) {
    fprintf(stderr, "Error: total number of MPI proc (%d) doesn't match product mx times my=%d\n",nProc,mx*my);
    exit(EXIT_FAILURE);
  }

  // initialize workspace memory (U, U2, ...)
  initHydroRun(&workspace);
  dt = compute_dt(&workspace, nStep%2 );

  // initialize boundaries
  make_boundaries_internal(workspace.U);
  make_boundaries_external(workspace.U);
  make_boundaries_internal(workspace.U2);
  make_boundaries_external(workspace.U2);

  // start computation
  if (myRank==0) printf("Start computation....\n");
  timerStart(&total_timer);

  catalyzer = new Catalyzer(MPI_COMM_WORLD);
  catalyzer->EnableSteering();
  catalyzer->SetInsituHostname("localhost");
  /* catalyzer->SetInsituPort(22223); */
  /* catalyzer->UsePythonPipeline("path_to/insitu.base.py"); */

  // Hydrodynamics solver loop
  while (t < tEnd && nStep < nStepmax) {

    // output
    if (nStep % nOutput == 0) {
      if (myRank==0) printf("Output results at time t=%2.9f step %d dt=%2.9f\n",t,nStep,dt);
      timerStart(&io_timer);
      saveVTK(workspace.U, nStep);
      timerStop(&io_timer);
    } // end output

    double d; int valid;
    catalyzer->GetSteeringValue(&d, &valid);
    // printf("[GetSteeringValue] %f %d\n", d, valid);

    // compute new dt
    timerStart(&dt_timer);
    dt = compute_dt(&workspace, nStep%2 );
    printf("dt value = %f\n", dt);

    // apply steering value
    if ( valid && d > 0 ) {
      dt = d;
      printf("[SteeringValue] New dt %f\n", dt);
    }
    timerStop(&dt_timer);
    
    // perform one step integration
    godunov_unsplit(&workspace, nStep, dt);

    // increase time
    nStep++;
    t+=dt;

  } // end solver loop

  delete catalyzer;

  // end of computation
  timerStop(&total_timer);

  // free workspace memory
  cleanupHydroRun(&workspace);

  // print monitoring information
  {
    double t_tot   = timerElapsed(&total_timer);
    double t_comp  = timerElapsed(&godunov_timer);
    double t_dt    = timerElapsed(&dt_timer);
    double t_bound = timerElapsed(&boundaries_timer);
    double t_io    = timerElapsed(&io_timer);
    
    double t_tot_max, t_comp_max, t_dt_max, t_bound_max, t_io_max;
    MPI_Reduce(&t_tot  , &t_tot_max  , 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
    MPI_Reduce(&t_comp , &t_comp_max , 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
    MPI_Reduce(&t_dt   , &t_dt_max   , 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
    MPI_Reduce(&t_bound, &t_bound_max, 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
    MPI_Reduce(&t_io   , &t_io_max   , 1, MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD);
    
    double t_tot_min, t_comp_min, t_dt_min, t_bound_min, t_io_min;
    MPI_Reduce(&t_tot  , &t_tot_min  , 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD);
    MPI_Reduce(&t_comp , &t_comp_min , 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD);
    MPI_Reduce(&t_dt   , &t_dt_min   , 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD);
    MPI_Reduce(&t_bound, &t_bound_min, 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD);
    MPI_Reduce(&t_io   , &t_io_min   , 1, MPI_DOUBLE, MPI_MIN, 0, MPI_COMM_WORLD);
    

    if (myRank == 0) {
      printf("total       time : %5.3f secondes\n",t_tot);
      printf("godunov     time : %5.3f secondes %5.2f%%\n",t_comp,t_comp/t_tot*100);
      printf("compute dt  time : %5.3f secondes %5.2f%%\n",t_dt,t_dt/t_tot*100);
      printf("boundaries  time : %5.3f secondes %5.2f%%\n",t_bound,t_bound/t_tot*100);
      printf("io          time : %5.3f secondes %5.2f%%\n",t_io,t_io/t_tot*100);
      printf("Perf             : %10.2f number of cell-updates/s\n",nStep*(mx*isize*my*jsize)/t_tot_max);
    }
  }

  MPI_Finalize();

  return EXIT_SUCCESS;

} // end main
